/* hrr_item_memory.pl */


/*
"Clean-up memory" for binary holographic 
reduced representation vectors.
*/


:- module( hrr_item_memory, [ hrr_item_memory_assert/2
                            , hrr_item_memory_retractall/0
                            , hrr_item_memory_retract_atom/1
                            , hrr_item_memory_retract_vector/1
                            , hrr_item_memory/2
                            , hrr_item_memory_listing/0
                            , hrr_item_memory_name/3
                            , hrr_item_memory_new_random_vector/3
                            , hrr_vector_clean/5
                            ] 
         ).


/*
This module implements a clean-up memory
for the binary reduced representation vector functions
implemented by the hrr_vectors module.

The memory is a set of pairs <Atom,Vector>.
Vectors in the memory are deemed to be
significant in some way. For example,
they may represent specific roles or
fillers. 

The most important operation using the
memory is the "clean up". This finds the
pair whose vector best matches a candidate
vector.

Exports:
  hrr_vector_clean( +Vector, +Dimension, -BestMatchAtom, -BestMatchVector, -Similarity ):
    This is the clean-up function. Finds the
    pair <BestMatchAtom,BestMatchVector> for which
    hrr_vector_dot(BestMatchVector,Vector) is
    greatest. That is, finds the most similar
    vector. Unifies Similarity with the
    similarity, which is the result of
    the dot product.

  hrr_item_memory_assert( +Atom, +Vector ):
    Adds the pair <Atom,Vector> to the memory.
    If a pair <Atom,_> is already there, deletes it
    first.

  hrr_item_memory_retractall:
    Deletes everything from the memory.

  hrr_item_memory_retract_atom( +Atom ):
    Deletes any pair whose atom is Atom.

  hrr_item_memory_retract_vector( +Vector ):
    Deletes any pair whose vector is Vector.

  hrr_item_memory( ?Atom, ?Vector ):
    Use for querying the memory. 

  hrr_item_memory_listing:
    Lists all atoms in the memory to the current
    output stream.

  hrr_item_memory_new_random_vector( +Atom, +Dimension, -Vector ):
    Generates a random vector of Dimension and
    adds it to the memory under the name Atom.

  hrr_item_memory_name( +Atom, +Dimension, -Vector ):
    If there is already a pair for Atom, returns
    its Vector. Otherwise, generates a new random
    vector of Dimension and adds it to the memory
    under the name Atom.

Implementation:
  The memory is implemented as clauses
  for the predicate hrr_item_memory(Atom,Vector).
  Clean-up searches these clauses, dotting
  each vector with its candidate and returning
  the pair with the greatest dot product.
*/


:- use_module( hrr(hrr_vectors) ).


:- dynamic hrr_item_memory/2.


hrr_item_memory_assert( Atom, Vec ) :-
  retractall( hrr_item_memory(Atom,_) ),
  asserta( hrr_item_memory(Atom,Vec) ).


hrr_item_memory_retractall :-
  retractall( hrr_item_memory(_,_) ).
 

hrr_item_memory_retract_atom(Atom) :-
  retractall( hrr_item_memory(Atom,_) ).


hrr_item_memory_retract_vector(Vec) :-
  retractall( hrr_item_memory(_,Vec) ).


hrr_item_memory_listing :-
  forall( hrr_item_memory(Atom,_Vector) 
        , ( write( Atom ), nl )
        ).


hrr_item_memory_name( Atom, _D, V ) :-
  hrr_item_memory( Atom, V ), !.

hrr_item_memory_name( Atom, D, V ) :-
  hrr_item_memory_new_random_vector( Atom, D, V ).


hrr_item_memory_new_random_vector( Atom, D, V ) :-
  hrr_vector_random( D, V ),
  assert( hrr_item_memory( Atom, V ) ).


hrr_vector_clean( V, D, AC, VC, BestSim ) :-
  nb_setval( hrr_item_memory_max, none ),
  nb_setval( hrr_item_memory_bestv, none ),
  nb_setval( hrr_item_memory_besta, none ),
  forall( hrr_item_memory(A,VA)
        , ( nb_getval( hrr_item_memory_max, Max ),
            hrr_vector_dot( V, VA, D, Sim ),
            ( Max = none -> 
                nb_setval( hrr_item_memory_max, Sim ), 
                nb_setval( hrr_item_memory_bestv, VA ),
                nb_setval( hrr_item_memory_besta, A ) 
            ; Sim > Max ->
                nb_setval( hrr_item_memory_max, Sim ), 
                nb_setval( hrr_item_memory_bestv, VA ),
                nb_setval( hrr_item_memory_besta, A )
            ;
              true
            )
          )
        ),
  nb_getval( hrr_item_memory_bestv, VC ),
  nb_getval( hrr_item_memory_besta, AC ),
  nb_getval( hrr_item_memory_max, BestSim ).
%
% This loops over the pairs, storing the current
% best vector in ..._bestv, its atom in ..._besta,
% and the similarity in ..._max.

