/* hrr_vectors.pl */


/*
Fundamental functions on binary holographic 
reduced representation vectors.
*/


:- module( hrr_vectors, [ hrr_vector_random/2
                        , hrr_vector_bind/3
                        , hrr_vector_probe/3
                        , hrr_vector_merge/2
                        , hrr_vector_dot/4
                        ]
         ).


/*
This module implements the holographic
reduced representation vector functions
described in Pentti Kanerva, "Dual Role of Analogy in the Design 
of a Cognitive Computer": in "Advances in Analogy Research:  
Integration of Theory and Data from the Cognitive, Computational, 
and Neural Sciences". Workshop. Sofia, Bulgaria, July 17-20, 1998.
http://faculty.cs.tamu.edu/choe/mirror/kanerva.ANALOGY98-kanerva.pdf .

These operate on binary vectors whose
elements are either 0 or 1. The functions
are: binding and unbinding of
vectors, both implemented as exclusive-or;
merging vectors; dot product, used to
measure similarity between vectors; and
generation of random vectors.

Exports:
  hrr_vector_random( +D, -V ):
    Returns in V a random vector of dimension D. 
    Each element has an equal probability of
    being 0 or 1.

  hrr_vector_bind( +V1, +V2, -V3 ):
     Binds V1 and V2 giving V3. This is
     implemented as exclusive-or. Kanerva
     writes it as 'x'-inside-circle
     in the above paper.
 
  hrr_vector_probe( +V1, +V2, -V3 ):
     Probes V1 with V2 giving V3.
     This is a synonym for hrr_vector_bind. 
     I distinguish them in case you want to
     be explicit about whether you're
     binding or probing.

  hrr_vector_merge( +ListOfVectors, -V ):
     Merges the vectors in the list giving
     V. Kanerva writes it as the result
     of a normalise (symbolised as [ ... ] )
     applied to the sum of vectors.
     So 
       hrr_vector_merge( [V1,V2], V )
     implements what Kanerva writes as
       V = [V1 + V2].
     Similarly,
       hrr_vector_merge( [V1,V2,V3], V )
     implements what Kanerva writes as
       V = [V1 + V2 + V3].
 
  hrr_vector_dot( +V1, +V2, +D, -S3 ):
     Forms the dot product of V1 and V2 giving 
     (scalar) S3. The dot product of V1 
     and V2 is S/D, where S is the sum over all elements
     V1i and V2i of (1 - V1i exclusive-or V2i),
     and D is the dimensionality
     of the vectors. I make D an
     argument, to avoid the predicate
     having to derive it by taking
     the length of one of its vectors.
     (I am not now sure where I got
     this definition of dot product:
     it isn't in the paper
     cited.)
*/ 


/*
Generating random vectors
=========================
*/


hrr_vector_random( 0, [] ) :- !.

hrr_vector_random( D, [ H | T ] ) :- 
  scalar_random( H ),
  Dd is D-1,
  hrr_vector_random( Dd, T ).


/* scalar_random( -S ):
     Returns a random scalar component 
     (equal choice of 0 or 1) in S.
*/ 
scalar_random( S ) :-
  S is random( 2 ).


/*
Binding and probing
===================
*/


hrr_vector_bind( V1, V2, V3 ) :- 
  maplist( scalar_bind, V1, V2, V3 ).


/* scalar_bind( +S1, +S2, -S3 ):
     Binds components S1 and S2 
     giving S3.
*/
scalar_bind( S1, S2, S3 ) :-
  S3 is S1 xor S2.


hrr_vector_probe( V1, V2, V3 ) :- 
   hrr_vector_bind( V1, V2, V3 ).


/*
Merging
=======
*/


hrr_vector_merge( Vecs, Merged ) :-
  length( Vecs, Len ),
  hrr_vector_merge( Vecs, Len, Merged ).


hrr_vector_merge( [[]|_], _Len, [] ) :- !.

hrr_vector_merge( Vecs, Len, [ MergedHead | MergedTail ] ) :-
  heads_and_tails( Vecs, Heads, Tails ),
  scalar_merge( Heads, Len, MergedHead ),
  hrr_vector_merge( Tails, Len, MergedTail ).
%
% As for the main predicate hrr_vector_merge/2,
% the first argument is a list of vectors to
% be merged. We iterate over this list,
% converting it into a list of their heads
% and a list of their tails. So we are splitting
% list [V1,...VN] into lists
% [V11,...VNi] and list [V1Tail,...VNTail].
% We do a scalar merge on the heads, thereby
% getting the first component in the merged 
% vector. We then recurse on the tails


/* heads_and_tails( +Vecs, -Heads, -Tails ):
     Vecs is a list of vectors [V1,...VN].
     Heads becomes a list of the first component
     of each, [V11,...VN1].
     Tails becomes a list of the remaining
     components, [V1Tail,...VNTail].
*/
heads_and_tails( [], [], [] ) :- !.

heads_and_tails( [[Head|Tail]|Lists], [Head|Heads], [Tail|Tails] ) :-
  heads_and_tails( Lists, Heads, Tails ).


/* scalar_merge( +Components, +Len, -Merged ):
     Given vectors V1 ... Vn, Components is a list
     of the components at the same position
     i in each, [V1i,...VNi].
     Len is the number of components.
     Merged will become the corresponding
     component in the merged vector. 
*/
scalar_merge( Components, Len, Merged ) :-
  count_0s( Components, N0s ),
  N1s is Len - N0s,
  majority_rule( N0s, N1s, Merged ).


/* majority_rule( +NumberOf0s, +NumberOf1s, -Component ):
     Implements Kanerva's "majority rule".
     Given vectors V1 ... VN, take components
     V1i ... VNi. 
     NumberOf0s and NumberOf1s are the number of
     components that are 0 or 1 respectively.
     Component will become 0 if there are
     more 0s than 1s, 1 if there are more 1s than 0s,
     and a randomly-chosen 0 or 1 if there are the
     same number of both. This last is the "tie-breaker".
*/
majority_rule( Ns, Ns, N ) :-
  !, N is random( 2 ).

majority_rule( N0s, N1s, 0 ) :-
  N0s > N1s, !.

majority_rule( _, _, 1 ).  
     

/* count_0s( +L, -N ):
     N is the number of 0's in
     list L. Uses the predicate
     below, which has an accumulator
     argument, to do so.
*/
count_0s( L, N ) :- 
  count_0s( L, 0, N ).


count_0s( [], SoFar, SoFar ) :- !.

count_0s( [0|T], SoFar, N ) :- !,
  SoFarPlus1 is SoFar + 1,
  count_0s( T, SoFarPlus1, N ).

count_0s( [_|T], SoFar, N ) :- !,
  count_0s( T, SoFar, N ).


/*
Dot product
===========
*/


hrr_vector_dot( V1, V2, D, S3 ) :-
  scalar_dots_and_sum( V1, V2, Sum ),
  S3 is Sum / D.


/* scalar_dot_and_sum( +V1, +V2, -Sum ):
     Sums the dot products of the
     components of V1 and V3 into Sum.
     Uses the auxiliary predicate below,
     which has an accumulator argument, 
     to do so.
*/
scalar_dots_and_sum( V1, V2, S3 ) :-
  scalar_dots_and_sum( V1, V2, 0, S3 ).


scalar_dots_and_sum( [], [], SoFar, SoFar ) :- !.

scalar_dots_and_sum( [S1|V1], [S2|V2], SoFar, Sum ) :- 
  scalar_dot( S1, S2, S1_Dot_S2 ),
  SoFarAndDot is SoFar + S1_Dot_S2,
  scalar_dots_and_sum( V1, V2, SoFarAndDot, Sum ).


/* scalar_dot( +S1, +S2, -S3 ):
     This dots corresponding components
     S1, S2 of vectors V1, V2, giving
     S3.
*/
scalar_dot( S1, S2, S3 ) :-
  S3 is 1 - (S1 xor S2).

