/* isv.pl */


/*
Interactive expression evaluator
for binary holographic 
reduced representation vectors.
*/


:- module( isv, [ (?=)/2
                , (isv)/2
                , (isv)/1
                , isv_dimension/1
                , isv_assert_dimension/1
                , isv_retract_dimension/0
                ]
         ).


/*
This module implements an analogue of
Prolog's is/2 predicate, for evaluating
binary reduced representation vector
expressions from the Prolog top-level
interpreter.

An expression E can be:
  A vector (which is a list);
  an atom, referring to a vector in the item memory;
  the term clean(E);
  the term merge(E1,...EN);
  the term E1 bind E2;
  the term E1 probe E2;
  a call F(E1,...EN).

Exports:
  isv( X, +Expr ):
    Expr is an expression as above.
    isv evaluates it. If X
    was an atom, isv then adds 
    the pair <X,ResultOfEvaluatingExpr>
    to the item memory, deleting
    any pair already there for the atom.
    This gives an easy way to calculate
    vectors (such as 'france' in Kanerva's
    example) and name them in the same step.
    If X was anything else, isv unifies it
    with the result of evaluating the expression.

    Note that if Expr is clean(...),
    isv assumes you want to know the cleaned
    vector's similarity. In this special
    case, it unifies its first argument
    with a pair MatchingAtom-Similarity.

  isv( +Expr ):
    Calls 
      X isv Expr
    and writes the result to the current
    output stream.

  ?=( +Vector, List ):
     List is either a list of pairs Slot-Value,
     or a single such pair.
     In each pair, Slot should be an atom.
     Value should be a variable.
     The predicate will do a probe and
     cleanup using each Slot, and return
     the name of the best match in Value.

  isv_assert_dimension( +D ):
     Assert that all subsquent pairs will
     be of dimension D. If this is
     different from the current dimension,
     and the item memory is not empty,
     you must also call hrr_item_memory_retractall,
     otherwise clean-ups using the vectors
     already in the memory will not work.

  isv_retract_dimension:
     Delete the dimension.

  isv_dimension( ?D ):
     Unify D with the dimension assumed for
     the vectors.
*/


:- use_module( hrr(hrr_vectors) ).
:- use_module( hrr(hrr_item_memory) ).


:- op( 700, xfx, user:isv ).
:- op( 700, fx, user:isv ).
:- op( 700, xfx, user:(?=) ).
:- op( 400, yfx, user:bind ).
:- op( 400, yfx, user:probe ).


:- dynamic isv_dimension/1.


isv_assert_dimension( D ) :-
  retractall( isv_dimension(_) ),
  assert( isv_dimension(D) ).


isv_retract_dimension :-
  retractall( isv_dimension(_) ).


?=( Expr, List ) :- 
  is_list( List ), !,
  hrr_eval( Expr, Vec ),
  '?=aux'( Vec, List ).
%
% Evaluate Expr here rather than
% in '?=aux', for efficiency.

?=( Expr, Atom-Value ) :- 
  atom( Atom ), !,
  ?=( Expr, [Atom-Value] ).


/* '?=aux'( +Vector, List ):
     List is a list of pairs Slot-Value.
     In each pair, Slot should be an atom.
     Value should be a variable.
     The predicate will do a probe and
     cleanup using each Slot, and return
     the name of the best match in Value.
*/ 
'?=aux'( _Vec, [] ) :- !.

'?=aux'( Vec, [Atom-Value|Rest] ) :-
  hrr_eval( Vec probe Atom, Probed ),
  isv_dimension( D ),
  hrr_vector_clean( Probed, D, Value, _ResultVector, _Similarity ),
  '?=aux'( Vec, Rest ).


isv( Expr ) :-
  isv( X, Expr ),
  write( X ), nl.


isv( ResultAtom, Expr ) :-
  atom( ResultAtom ), !,
  hrr_eval( Expr, ResultVector ),
  hrr_item_memory_retract_atom( ResultAtom ),
  hrr_item_memory_assert( ResultAtom, ResultVector ).
%
% This is the clause that names a result:
% ResultAtom is an atom, so we add the pair
% <ResultAtom,ResultVector> to the item memory.

isv( ResultAtom - Similarity, clean(E) ) :- !,
  hrr_eval( E, EEvald ),
  isv_dimension( D ),
  hrr_vector_clean( EEvald, D, ResultAtom, _ResultVector, Similarity ).
%
% This is a special case for 'clean'.
% We return the matching vector _and_
% the similarity in the first argument,
% so that the user can see how good the
% match was.

isv( Result, Expr ) :-
  hrr_eval( Expr, Result ).
%
% The "standard" case. Just call
% hrr_eval.


/* hrr_eval( +Expr, -Vec ):
     This is the guts of isv.
     Expr is an expression as defined
     at the start of this file. Vec
     becomes the result of
     evaluating it. hrr_eval
     recursively splits
     Expr into subexpressions
     and evaluates those, combining
     the results.
*/
hrr_eval( Vec, Vec ) :-
  is_list( Vec ), !.

hrr_eval( Atom, ResultVector ) :-
  atom( Atom ), !,
  isv_dimension( D ),
  hrr_item_memory_name( Atom, D, ResultVector ).

hrr_eval( clean(E), ECleaned ) :- !,
  hrr_eval( E, EEvald ),
  isv_dimension( D ),
  hrr_vector_clean( EEvald, D, _Atom, ECleaned, _Similarity ).

hrr_eval( Merge, Result ) :- 
  Merge =.. [ merge | Es ], !,
  hrr_eval_args( Es, EsEvald ),
  hrr_vector_merge( EsEvald, Result ).

hrr_eval( E1 bind E2, Result ) :- !,
  hrr_eval( E1, E1Evald ),
  hrr_eval( E2, E2Evald ),
  hrr_vector_bind( E1Evald, E2Evald, Result ).

hrr_eval( E1 probe E2, Result ) :- !,
  hrr_eval( E1, E1Evald ),
  hrr_eval( E2, E2Evald ),
  hrr_vector_probe( E1Evald, E2Evald, Result ).

hrr_eval( Expr, Result ) :-
  Expr =.. [ F | Args ],
  hrr_eval_args( Args, EvaldArgs ),
  append( EvaldArgs, [Result], All ),
  Goal =.. [ F | All ],
  call( Goal ).


/* hrr_eval_args( +Args, -EvaldArgs ):
     Maps hrr_eval over a list Args of
     expressions, returning a list
     of evaluated expressions in
     EvaldArgs.
*/
hrr_eval_args( Args, Result ) :-
  maplist( hrr_eval, Args, Result ).


